{****************************************************************************}
{                      Adapted for Data Master 2000                          }
{          portions of code used from PASLIB library by Will Polik           }
{****************************************************************************}
unit LMFit;
{$R-}
interface

uses Common;

const
  MaxMatSize = {14}MaxCols;     {maximum number of fitting parameters}
  MaxVectorSize = {6}MaxCols;   {maximum dimension of independent variables}
  MaxDataSize = MaxCols;        {maximum number of data points}

type
  DataArray = Array[1..MaxDataSize] of TReal;
  DataPointer = ^DataArray;
  Vector = Array[1..MaxVectorSize] of TReal;
  VectorArray = Array[1..MaxDataSize] of Vector;
  VectorPointer = ^VectorArray;
  FitArray  = Array[1..MaxMatSize] of TReal;
  InvMatrix = Array[1..MaxMatSize,1..MaxMatSize] of TReal;
  TFunc=function(X: TReal; A: FitArray; NParam: Integer) : TReal;
  TDeriv=procedure(X: TReal; A,SigA: FitArray; Nparam: Integer;
    var Deriv: FitArray);            // don't calculate Deriv[i] if SigA[i]<0!
  TProgress=procedure;
  TWeightType=(wtEqual,wtRelative,wtAbsolute);
  TFitResult=(frParamConv,frChiConv,frMaxIter,frFitError,frOther);

procedure CurveFit(X,Y,W: DataPointer; Npts,Nparam: Integer; var A,SigmaA:
  FitArray; var ChiSqR: TReal; var Yfit: DataPointer; var Niter: Integer;
  ParDel,ChiDel: TReal; WghtTyp: TWeightType; fu: TFunc; fd: TDeriv;
  fp: TProgress; var Code: TFitResult);

implementation

uses Classes, SysUtils;

procedure SymMatInv(var Matrix: InvMatrix; N:Integer; var Determinant: TReal);

{Inverts a symmetric matrix and calculates its determinant

 by Will Polik, UC Berkeley, 5/6/86

 Adapted from Bevington, Data Reduction and Error Analysis for the Physical
 Sciences, McGraw-Hill, New York, 1969, p. 302

 Assumes:
   const
     MaxMatSize = ??;
   type
     InvMatrix = Array[1..MaxMatSize,1..MaxMatSize] of TReal;

 Parameters:
   Matrix      - Input matrix which is replaced by its inverse
   N           - Degree of matrix
   Determinant - Determinant of Input Matrix}

label
  TryAgain;

var
  I,J,K,L   : Integer;
  Amax,Save : TReal;
  IK,JK     : Array[1..MaxMatSize] of Integer;

begin {of procedure SymMatInv}
if N > MaxMatSize
  then Exception.CreateFmt('ERROR: N=%d > MaxMatSize=%d in SymMatInv',
  [N,MaxMatSize]);
Determinant := 1.0;
for K := 1 to N do
  begin
  {find largest element in rest of matrix}
  Amax := 0;
  TryAgain:
  for I := K to N do
    begin
    for J := K to N do
      if Abs(Amax) <= Abs(Matrix[I,J])
      then
        begin
        Amax := Matrix[I,J];
        IK[K] := I;
        JK[K] := J;
        end;
    end;
  {Interchange rows and columns to put Amax in Matrix[K,K]}
  if Amax = 0
  then
    begin
    Determinant := 0;
    exit;
    end;
  I := IK[K];
  if K > I then goto TryAgain;
  if I > K
  then for J := 1 to N do
    begin
    Save := Matrix[K,J];
    Matrix[K,J] := Matrix[I,J];
    Matrix[I,J] := -Save;
    end;
  J := JK[K];
  if K > J then goto TryAgain;
  if J > K
  then for I := 1 to N do
    begin
    Save := Matrix[I,K];
    Matrix[I,K] := Matrix[I,J];
    Matrix[I,J] := -Save;
    end;
  {Accumulate elements of inverse matrix}
  for I := 1 to N do
    if I <> K
    then Matrix[I,K] := -Matrix[I,K]/Amax;
  for I := 1 to N do
    begin
    for J := 1 to N do
      if (I <> K) and (J <> K)
      then Matrix[I,J] := Matrix[I,J] + Matrix[I,K] * Matrix[K,J];
    end;
  for J := 1 to N do
    if J <> K
    then Matrix[K,J] := Matrix[K,J]/Amax;
  Matrix[K,K] := 1.0/Amax;
  Determinant := Determinant * Amax;
  end;
{Restore ordering of matrix}
for L := 1 to N do
  begin
  K := N - L + 1;
  J := IK[K];
  if J > K
  then for I := 1 to N do
    begin
    Save := Matrix[I,K];
    Matrix[I,K] := -Matrix[I,J];
    Matrix[I,J] := Save;
    end;
  I := JK[K];
  if I > K
  then for J := 1 to N do
    begin
    Save := Matrix[K,J];
    Matrix[K,J] := -Matrix[I,J];
    Matrix[I,J] := Save;
    end;
  end;
end; {of procedure SymMatInv}

procedure CurveFit(X,Y,W: DataPointer; Npts,Nparam: Integer; var A,SigmaA:
  FitArray; var ChiSqR: TReal; var Yfit: DataPointer; var Niter: Integer;
  ParDel,ChiDel: TReal; WghtTyp: TWeightType; fu: TFunc; fd: TDeriv;
  fp: TProgress; var Code: TFitResult);

{Makes a least-squares fit to a non-linear function with a
 linearization of the fitting function using Marquardt's algorithm:

   X       - Pointer to array of independent variable data points
   Y       - Pointer to array of dependent variable data points
   W       - Pointer to array of relative weights of data points
   Npts    - Number of pairs of data points
   Nparam  - Number of fitting parameters (variable and fixed)
   A       - Array of fitting parameters
   SigmaA  - Array of standard deviations for parameters A
            (Set SigmaA[j] < 0 if A[j] is fixed)
   ChiSqr  - Reduced chi squared for fit
   Yfit    - Pointer to array of calculated values of Y
   Niter   - Number of iterations completed
   ParDel  - Convergence threshold for parameter changes (ex: 1E-7)
   ChiDel  - Convergence threshold for Chi Square changes (ex: 1E-4)
   WghtTyp - Type of weighting:  0 = Equal weighting
                                 1 = Relative weighting
                                 2 = Absolute weighting (W^[] := 1/Sqr(Sigma))
   ---- Added by RRR:----
   fu - calculates function value for given X and parameters;
   fd - calculates function partial derivatives for given X and parameters;
   fp - may display the results of each iteration to the screen.

 Assumes:
  const
    MaxMatSize = ??;
  type
    DataArray = Array[1..???] of TReal;
    DataPointer = ^DataArray;
    FitArray  = Array[1..MaxMatSize] of TReal;
    InvMatrix = Array[1..MaxMatSize,1..MaxMatSize] of TReal;
  $I SYMATINV.PRO
  procedure FunctnOnce(A: FitArray; Nparam: Integer);
  ...Calculate X-independent terms (global variables) used by Functn...
  procedure FDerivOnce(A: FitArray; Nparam: Integer);
  ...Calculate X-independent terms (global variables) used by FDeriv...
  function Functn(X: TReal; A : FitArray; Nparam: Integer) : TReal;
  ...Return Functn evaluated at X with parameters A...
  procedure FDeriv(X: TReal; A : FitArray; Nparam: Integer;
                   var Deriv : FitArray);
  ...Evaluate partial derivatives of Functn w.r.t. parameters A at X...
  ...No need to evaluate Deriv[j] when SigmaA[j] < 0...}

procedure CurFit(X,Y,W: DataPointer; Npts,Nparam,Nfixed,Nvar: Integer; var A,
  SigmaA: FitArray; var ChiSqR: TReal; var Yfit: DataPointer;
  var Flamda: TReal; var FitError: boolean; WghtTyp: TWeightType);

{Adapted from Bevington, Data Reduction and Error Analysis for the Physical
 Sciences, McGraw-Hill, New York, 1969, p. 237.

 Improvements by William F. Polik, U.C. Berkeley, 1986:
   1) Allow for arbitrary relative weighting of points, i.e., weights do
      not have to be SigmaY.
   2) Allow for fixing variables; fixed variable is indicated by a negative
      value for SigmaA[j].
   3) Protection has been added against Flamda underflow or overflow.
   4) The option to print out the results of each iteration has been added.
   5) A driver routine that monitors for convergence has been added.
      Convergence will occur when every parameter or chi squared changes by
      less than a factor specified by ParDel or ChiDel.
   6) Functn and FDeriv calculation has been made more efficient by allowing
      all X independent terms to be evaluated only once each iteration in
      FunctnOnce and FDerivOnce; intermediate results must be stored in global
      variables which are then referenced by Functn and FDeriv.  If this
      feature is not desired, FunctnOnce and FDerivOnce should be written as
      dummy "do nothing" procedures.
   7) An optional procedure (commented out) allows for checking each parameter
      value to see if it has taken on an "erroneous" value after each
      iteration; if so, it may be reset to a "safe" value and then fixed.  See
      the procedure CheckFit.
   8) Equal, relative, or absolute weighting is provided.}

label
  TryAgain;

var
  i,                         {Index for points}
  j,k,                       {Index for all parameters}
  jvar,kvar : Integer;       {Index for variable parameters}
  Alpha,                     {First partial derivative matrix}
  CurvMat   : InvMatrix;     {Normalized curvature matrix, error matrix}
  Beta      : Array[1..MaxMatSize] of TReal;
  Deriv,                     {First partial derivatives of Functn at X^[i]}
  B         : FitArray;      {Next guess at parameters}
  Determ    : TReal;         {Matrix determinant - not used}
  ChiSqR1   : TReal;         {Original reduced chi squared}
  Wtot      : TReal;         {Total weights}

begin {of procedure CurFit}
for jvar := 1 to Nvar do               {Zero Alpha, Beta}
  begin
  for kvar := 1 to Nvar do
    Alpha[jvar,kvar] := 0;
  Beta[jvar] := 0;
  end;
// FunctnOnce(A,Nparam); not used by RRR
// FDerivOnce(A,Nparam);
for i := 1 to Npts do                  {Evaluate Alpha and Beta}
  begin
  Yfit^[i] := Fu(X^[i],A,NParam);
  FD(X^[i],A,SigmaA,Nparam,Deriv);
  jvar := 1;
  for j := 1 to Nparam do
    begin
    if SigmaA[j] >= 0 then
      begin
      Beta[jvar] := Beta[jvar] + W^[i]*(Y^[i]-Yfit^[i])*Deriv[j];
      kvar := 1;
      for k := 1 to j do
        begin
        if SigmaA[k] >= 0 then
          begin
          Alpha[jvar,kvar] := Alpha[jvar,kvar] + W^[i]*Deriv[j]*Deriv[k];
          kvar := kvar + 1;
          end;
        end;
      jvar := jvar + 1;
      end;
    end;
  end;
for jvar := 1 to Nvar do
  for kvar := 1 to (jvar-1) do
    Alpha[kvar,jvar] := Alpha[jvar,kvar];
ChiSqR1 := 0;                          {Evaluate Chi square at starting point}
Wtot := 0;
for i := 1 to Npts do
  begin
  ChiSqR1 := ChiSqR1 + W^[i]*Sqr(Y^[i]-Yfit^[i]);
  if WghtTyp in [wtEqual,wtRelative]
  then Wtot := Wtot + W^[i]
  else Wtot := Wtot + 1;
  end;
ChiSqR1 := ChiSqR1*(Npts/Wtot)/(Npts-Nvar);
TryAgain:;
for jvar := 1 to Nvar do               {Invert modified curvature matrix to}
  begin                                {find new parameters}
  for kvar := 1 to Nvar do
    CurvMat[jvar,kvar] := Alpha[jvar,kvar]/
                          Sqrt(Alpha[jvar,jvar]*Alpha[kvar,kvar]);
  CurvMat[jvar,jvar] := 1.0 + Flamda;
  end;
SymMatInv(CurvMat,Nvar,Determ);
for j := 1 to Nparam do                {Find new parameters}
  B[j] := A[j];
jvar := 1;
for j := 1 to NParam do
  begin
  if SigmaA[j] >= 0 then
    begin
    kvar := 1;
    for k := 1 to NParam do
      begin
      if SigmaA[k] >= 0 then
        begin
        B[j] := B[j] + Beta[kvar]*CurvMat[jvar,kvar]/
                       Sqrt(Alpha[jvar,jvar]*Alpha[kvar,kvar]);
        kvar := kvar + 1;
        end;
      end;
    jvar := jvar + 1;
    end;
  end;
// FunctnOnce(B,Nparam);
for i := 1 to Npts do                  {If chi squared increased, increase}
  Yfit^[i] := Fu(X^[i],B,Nparam);      {Flambda and try again}
ChiSqR := 0;
for i := 1 to Npts do
  ChiSqR := ChiSqR + W^[i]*Sqr(Y^[i]-Yfit^[i]);
ChiSqR := ChiSqR*(Npts/Wtot)/(Npts-Nvar);
if ChiSqR > ChiSqR1
then
  begin
  Flamda := 10.0 * Flamda;
  if Flamda > 10E+35
  then
    begin
    FitError := true;
    exit;
    end;
  Goto TryAgain;
  end;
jvar := 1;
for j := 1 to Nparam do                {Chi squared decreased, so}
  begin                                {set returned parameters and}
  if SigmaA[j] >= 0 then               {uncertainties}
    begin
    A[j] := B[j];
    if WghtTyp in [wtEqual,wtRelative]
    then SigmaA[j] := Sqrt(Abs(ChiSqR*CurvMat[jvar,jvar]*
                          (Wtot/Npts)/Alpha[jvar,jvar]))
    else SigmaA[j] := Sqrt(Abs(CurvMat[jvar,jvar]/Alpha[jvar,jvar]));
    jvar := jvar + 1;
    end;
  end;
Flamda := Flamda/10.0;
if Flamda < 10E-35
then Flamda := 10E-35;
FitError := false;
end; {of procedure CurFit}

var
  Nfixed,                    {Number of fixed parameters}
  Nvar           : Integer;  {Number of variable parameters}
  i              : Integer;  {Index}
  Flamda         : TReal;    {Proportion of gradient search included}
  FitError       : boolean;  {Error code returned by Curfit}
  Aold           : FitArray; {Previous set of parameters}
  ChiSqRold      : TReal;    {Previous reduced chi squared}
  ParamConverged,
  ChiConverged   : Boolean;  {Flags for convergence}
  MaxIter        : integer;  // added by RRR

(*
procedure CheckFit;
{If any parameters are fixed or freed, Nfixed and Nvar must be recalculated;
 also, if range checking is off (default!), error checking should be repeated,
 and another iteration should be initiated}
var
  i,j : Integer;
begin {of procedure CheckFit}
for i := 1 to Nparam do
  begin
  if ...error condition in A[i]...
  then
    begin
    ...reset the A[i] to a fixed value (some global variable representing
       the initial guess for A[i]...
    SigmaA[i] := -1.0;                     {Fix A[i]}
    ChiSqROld := 1.1 * ChiSqR;             {Insure another interation}
    Nfixed := 0;                           {Find number of fixed and}
    for j := 1 to Nparam do                {variable parameters}
      if SigmaA[j] < 0 then Nfixed := Nfixed + 1;
    Nvar := Nparam - Nfixed;
    if Nvar > MaxMatSize                   {Error checking}
    then
      begin
      WriteLn('ERROR: Nvar = ',Nvar,' > MaxMatSize = ',MaxMatSize,' in CheckFit');
      halt;
      end;
    if Nvar >= Npts                        {Error checking}
    then
      begin
      WriteLn('ERROR: Nvar = ',Nvar,' >= Npts = ',Npts,' in CheckFit');
      halt;
      end;
    end;
  end;
end; {of procedure CheckFit}
*)

begin {of procedure CurveFit}
{!!! RRR: check values} ParDel:=Abs(ParDel); ChiDel:=Abs(ChiDel);
if WghtTyp = wtEqual
then
  for i := 1 to NPts do
    W^[i] := 1;
Nfixed := 0;                           {Find number of fixed and}
for i := 1 to Nparam do                {variable parameters}
  if SigmaA[i] < 0 then Nfixed := Nfixed + 1;
Nvar := Nparam - Nfixed;

if Nvar > MaxMatSize                   {Error checking}
then Exception.CreateFmt('ERROR: Nvar=%d > MaxMatSize=%d in CurveFit',
  [Nvar,MaxMatSize]);
if Nvar >= Npts                        {Error checking}
then Exception.CreateFmt('ERROR: Nvar=%d >= Npts=%d in CurveFit',
  [Nvar,Npts]);

Flamda := 0.001;                       {Try to fit first time}
CurFit(X,Y,W,Npts,Nparam,Nfixed,Nvar,A,SigmaA,ChiSqR,Yfit,
  Flamda,FitError,WghtTyp);

MaxIter:=NIter; ///
Niter := 1;
if Assigned(fp) then fp;
for i := 1 to Nparam do                {Save convergence critera values}
  Aold[i] := A[i];
ChiSqRold := ChiSqR;
(*
CheckFit;
*)
ParamConverged := False;               {Continue until curve is fit}
ChiConverged := False;
While (ParamConverged=False) and (ChiConverged=False) and (not FitError)
 and (Niter<=MaxIter) {added by RRR} do
  begin
  CurFit(X,Y,W,Npts,Nparam,Nfixed,Nvar,A,SigmaA,ChiSqR,Yfit,
    Flamda,FitError,WghtTyp);
(*
  CheckFit;
*)
  Niter := Niter + 1;
  if Assigned(fp) then fp;
  if not FitError
  then
    begin
    ParamConverged := ParDel>0{True};  // modified by RRR
    i := 1;                            {Check for parameter convergence}
    While (i <= Nparam) and (ParamConverged = True) do
      begin
      if Abs(Aold[i]-A[i]) > Abs(ParDel*A[i])
      then ParamConverged := False;
      i := i + 1;
      end;
    if (ChiSqR >= ChiSqRold) or (ChiDel=0){!} or      {Check for chi squared}
       (Abs(ChiSqRold-ChiSqR) > Abs(ChiDel*ChiSqR))   {convergence}
    then ChiConverged := False
    else ChiConverged := True;
    for i := 1 to Nparam do
      Aold[i] := A[i];
    ChiSqRold := ChiSqR;
    end;
  end;
if ParamConverged then                 {Set result code}
  Code := frParamConv
else if ChiConverged then
  Code := frChiConv
else if FitError then
  Code := frFitError
else if NIter>=MaxIter then //
  Code:=frMaxIter           //
else Code := frOther;
end; {of procedure CurveFit}

end.
